"""
Managing Gateway Groups and interactions with multiple channels.

(c) 2008-2009, Holger Krekel and others
"""

import os, sys, atexit
import time
import execnet
from execnet.threadpool import WorkerPool

from execnet import XSpec
from execnet import gateway, gateway_io, gateway_bootstrap
from execnet.gateway_base import queue, reraise, trace, TimeoutError

NO_ENDMARKER_WANTED = object()

class Group:
    """ Gateway Groups. """
    defaultspec = "popen"
    def __init__(self, xspecs=()):
        """ initialize group and make gateways as specified. """
        # Gateways may evolve to become GC-collectable
        self._gateways = []
        self._autoidcounter = 0
        self._gateways_to_join = []
        for xspec in xspecs:
            self.makegateway(xspec)
        atexit.register(self._cleanup_atexit)

    def __repr__(self):
        idgateways = [gw.id for gw in self]
        return "<Group %r>" %(idgateways)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._gateways[key]
        for gw in self._gateways:
            if gw == key or gw.id == key:
                return gw
        raise KeyError(key)

    def __contains__(self, key):
        try:
            self[key]
            return True
        except KeyError:
            return False

    def __len__(self):
        return len(self._gateways)

    def __iter__(self):
        return iter(list(self._gateways))

    def makegateway(self, spec=None):
        """create and configure a gateway to a Python interpreter.
        The ``spec`` string encodes the target gateway type
        and configuration information. The general format is::

            key1=value1//key2=value2//...

        If you leave out the ``=value`` part a True value is assumed.
        Valid types: ``popen``, ``ssh=hostname``, ``socket=host:port``.
        Valid configuration::

            id=<string>     specifies the gateway id
            python=<path>   specifies which python interpreter to execute
            chdir=<path>    specifies to which directory to change
            nice=<path>     specifies process priority of new process
            env:NAME=value  specifies a remote environment variable setting.

        If no spec is given, self.defaultspec is used.
        """
        if not spec:
            spec = self.defaultspec
        if not isinstance(spec, XSpec):
            spec = XSpec(spec)
        self.allocate_id(spec)
        if spec.via:
            assert not spec.socket
            master = self[spec.via]
            channel = master.remote_exec(gateway_io)
            channel.send(vars(spec))
            io = gateway_io.RemoteIO(channel)
            gw = gateway_bootstrap.bootstrap(io, spec)
        elif spec.popen or spec.ssh:
            io = gateway_io.create_io(spec)
            gw = gateway_bootstrap.bootstrap(io, spec)
        elif spec.socket:
            from execnet import gateway_socket
            io = gateway_socket.create_io(spec, self)
            gw = gateway_bootstrap.bootstrap(io, spec)
        else:
            raise ValueError("no gateway type found for %r" % (spec._spec,))
        gw.spec = spec
        self._register(gw)
        if spec.chdir or spec.nice or spec.env:
            channel = gw.remote_exec("""
                import os
                path, nice, env = channel.receive()
                if path:
                    if not os.path.exists(path):
                        os.mkdir(path)
                    os.chdir(path)
                if nice and hasattr(os, 'nice'):
                    os.nice(nice)
                if env:
                    for name, value in env.items():
                        os.environ[name] = value
            """)
            nice = spec.nice and int(spec.nice) or 0
            channel.send((spec.chdir, nice, spec.env))
            channel.waitclose()
        return gw

    def allocate_id(self, spec):
        """ allocate id for the given xspec object. """
        if spec.id is None:
            id = "gw" + str(self._autoidcounter)
            self._autoidcounter += 1
            if id in self:
                raise ValueError("already have gateway with id %r" %(id,))
            spec.id = id

    def _register(self, gateway):
        assert not hasattr(gateway, '_group')
        assert gateway.id
        assert id not in self
        self._gateways.append(gateway)
        gateway._group = self

    def _unregister(self, gateway):
        self._gateways.remove(gateway)
        self._gateways_to_join.append(gateway)

    def _cleanup_atexit(self):
        trace("=== atexit cleanup %r ===" %(self,))
        self.terminate(timeout=1.0)

    def terminate(self, timeout=None):
        """ trigger exit of member gateways and wait for termination
        of member gateways and associated subprocesses.  After waiting
        timeout seconds try to to kill local sub processes of popen-
        and ssh-gateways.  Timeout defaults to None meaning
        open-ended waiting and no kill attempts.
        """

        while self:
            from execnet.threadpool import WorkerPool
            vias = {}
            for gw in self:
                if gw.spec.via:
                    vias[gw.spec.via] = True
            for gw in self:
                if gw.id not in vias:
                    gw.exit()

            def join_wait(gw):
                gw.join()
                gw._io.wait()
            def kill(gw):
                trace("Gateways did not come down after timeout: %r" % gw)
                gw._io.kill()

            safe_terminate(timeout, [
                (lambda: join_wait(gw), lambda: kill(gw))
                for gw in self._gateways_to_join])
            self._gateways_to_join[:] = []

    def remote_exec(self, source, **kwargs):
        """ remote_exec source on all member gateways and return
            MultiChannel connecting to all sub processes.
        """
        channels = []
        for gw in self:
            channels.append(gw.remote_exec(source, **kwargs))
        return MultiChannel(channels)

class MultiChannel:
    def __init__(self, channels):
        self._channels = channels

    def __len__(self):
        return len(self._channels)

    def __iter__(self):
        return iter(self._channels)

    def __getitem__(self, key):
        return self._channels[key]

    def __contains__(self, chan):
        return chan in self._channels

    def send_each(self, item):
        for ch in self._channels:
            ch.send(item)

    def receive_each(self, withchannel=False):
        assert not hasattr(self, '_queue')
        l = []
        for ch in self._channels:
            obj = ch.receive()
            if withchannel:
                l.append((ch, obj))
            else:
                l.append(obj)
        return l

    def make_receive_queue(self, endmarker=NO_ENDMARKER_WANTED):
        try:
            return self._queue
        except AttributeError:
            self._queue = queue.Queue()
            for ch in self._channels:
                def putreceived(obj, channel=ch):
                    self._queue.put((channel, obj))
                if endmarker is NO_ENDMARKER_WANTED:
                    ch.setcallback(putreceived)
                else:
                    ch.setcallback(putreceived, endmarker=endmarker)
            return self._queue


    def waitclose(self):
        first = None
        for ch in self._channels:
            try:
                ch.waitclose()
            except ch.RemoteError:
                if first is None:
                    first = sys.exc_info()
        if first:
            reraise(*first)



def safe_terminate(timeout, list_of_paired_functions):
    workerpool = WorkerPool(len(list_of_paired_functions)*2)

    def termkill(termfunc, killfunc):
        termreply = workerpool.dispatch(termfunc)
        try:
            termreply.get(timeout=timeout)
        except IOError:
            killfunc()

    replylist = []
    for termfunc, killfunc in list_of_paired_functions:
        reply = workerpool.dispatch(termkill, termfunc, killfunc)
        replylist.append(reply)
    for reply in replylist:
        reply.get()


default_group = Group()
makegateway = default_group.makegateway






def admin_command(sudo, command):
    """
    If sudo is needed, make sure the command is prepended
    correctly, otherwise return the command as it came.

    :param sudo: A boolean representing the intention of having a sudo command
                (or not)
    :param command: A list of the actual command to execute with Popen.
    """
    if sudo:
        command.insert(0, 'sudo')
    return command



#! /usr/bin/env python

"""
    start socket based minimal readline exec server

    it can exeuted in 2 modes of operation

    1. as normal script, that listens for new connections

    2. via existing_gateway.remote_exec (as imported module)

"""
# this part of the program only executes on the server side
#

progname = 'socket_readline_exec_server-1.2'

import sys, socket, os
try:
    import fcntl
except ImportError:
    fcntl = None

debug = 0

if debug: #  and not os.isatty(sys.stdin.fileno()):
    f = open('/tmp/execnet-socket-pyout.log', 'w')
    old = sys.stdout, sys.stderr
    sys.stdout = sys.stderr = f

def print_(*args):
    print(" ".join(str(arg) for arg in args))

if sys.version_info > (3, 0):
    exec("""def exec_(source, locs):
    exec(source, locs)""")
else:
    exec("""def exec_(source, locs):
    exec source in locs""")

def exec_from_one_connection(serversock):
    print_(progname, 'Entering Accept loop', serversock.getsockname())
    clientsock,address = serversock.accept()
    print_(progname, 'got new connection from %s %s' % address)
    clientfile = clientsock.makefile('rb')
    print_("reading line")
    # rstrip so that we can use \r\n for telnet testing
    source = clientfile.readline().rstrip()
    clientfile.close()
    g = {'clientsock' : clientsock, 'address' : address}
    source = eval(source)
    if source:
        co = compile(source+'\n', source, 'exec')
        print_(progname, 'compiled source, executing')
        try:
            exec_(co, g)
        finally:
            print_(progname, 'finished executing code')
            # background thread might hold a reference to this (!?)
            #clientsock.close()

def bind_and_listen(hostport):
    if isinstance(hostport, str):
        host, port = hostport.split(':')
        hostport = (host, int(port))
    serversock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # set close-on-exec
    if hasattr(fcntl, 'FD_CLOEXEC'):
        old = fcntl.fcntl(serversock.fileno(), fcntl.F_GETFD)
        fcntl.fcntl(serversock.fileno(), fcntl.F_SETFD, old | fcntl.FD_CLOEXEC)
    # allow the address to be re-used in a reasonable amount of time
    if os.name == 'posix' and sys.platform != 'cygwin':
        serversock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    serversock.bind(hostport)
    serversock.listen(5)
    return serversock

def startserver(serversock, loop=False):
    try:
        while 1:
            try:
                exec_from_one_connection(serversock)
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                if debug:
                    import traceback
                    traceback.print_exc()
                else:
                    excinfo = sys.exc_info()
                    print_("got exception", excinfo[1])
            if not loop:
                break
    finally:
        print_("leaving socketserver execloop")
        serversock.shutdown(2)

if __name__ == '__main__':
    import sys
    if len(sys.argv)>1:
        hostport = sys.argv[1]
    else:
        hostport = ':8888'
    serversock = bind_and_listen(hostport)
    startserver(serversock, loop=False)
elif __name__=='__channelexec__':
    bindname = channel.receive()
    sock = bind_and_listen(bindname)
    port = sock.getsockname()
    channel.send(port)
    startserver(sock)



#! /usr/bin/env python
"""
a remote python shell

for injection into startserver.py
"""
import sys, os, socket, select

try:
    clientsock
except NameError:
    print("client side starting")
    import sys
    host, port  = sys.argv[1].split(':')
    port = int(port)
    myself = open(os.path.abspath(sys.argv[0]), 'rU').read()
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((host, port))
    sock.sendall(repr(myself)+'\n')
    print("send boot string")
    inputlist = [ sock, sys.stdin ]
    try:
        while 1:
            r,w,e = select.select(inputlist, [], [])
            if sys.stdin in r:
                line = raw_input()
                sock.sendall(line + '\n')
            if sock in r:
                line = sock.recv(4096)
                sys.stdout.write(line)
                sys.stdout.flush()
    except:
        import traceback
        print(traceback.print_exc())

    sys.exit(1)

print("server side starting")
# server side
#
from traceback import print_exc
from threading import Thread

class promptagent(Thread):
    def __init__(self, clientsock):
        Thread.__init__(self)
        self.clientsock = clientsock

    def run(self):
        print("Entering thread prompt loop")
        clientfile = self.clientsock.makefile('w')

        filein = self.clientsock.makefile('r')
        loc = self.clientsock.getsockname()

        while 1:
            try:
                clientfile.write('%s %s >>> ' % loc)
                clientfile.flush()
                line = filein.readline()
                if len(line)==0: raise EOFError("nothing")
                #print >>sys.stderr,"got line: " + line
                if line.strip():
                    oldout, olderr = sys.stdout, sys.stderr
                    sys.stdout, sys.stderr = clientfile, clientfile
                    try:
                        try:
                            exec(compile(line + '\n','<remote pyin>', 'single'))
                        except:
                            print_exc()
                    finally:
                        sys.stdout=oldout
                        sys.stderr=olderr
                clientfile.flush()
            except EOFError:
                e = sys.exc_info()[1]
                sys.stderr.write("connection close, prompt thread returns")
                break
                #print >>sys.stdout, "".join(apply(format_exception,sys.exc_info()))

        self.clientsock.close()

prompter = promptagent(clientsock)
prompter.start()
print("promptagent - thread started")



"""
some deprecated calls

(c) 2008-2009, Holger Krekel and others
"""
import execnet

def PopenGateway(python=None):
    """ instantiate a gateway to a subprocess
        started with the given 'python' executable.
    """
    APIWARN("1.0.0b4", "use makegateway('popen')")
    spec = execnet.XSpec("popen")
    spec.python = python
    return execnet.default_group.makegateway(spec)

def SocketGateway(host, port):
    """ This Gateway provides interaction with a remote process
        by connecting to a specified socket.  On the remote
        side you need to manually start a small script
        (py/execnet/script/socketserver.py) that accepts
        SocketGateway connections or use the experimental
        new_remote() method on existing gateways.
    """
    APIWARN("1.0.0b4", "use makegateway('socket=host:port')")
    spec = execnet.XSpec("socket=%s:%s" %(host, port))
    return execnet.default_group.makegateway(spec)

def SshGateway(sshaddress, remotepython=None, ssh_config=None):
    """ instantiate a remote ssh process with the
        given 'sshaddress' and remotepython version.
        you may specify an ssh_config file.
    """
    APIWARN("1.0.0b4", "use makegateway('ssh=host')")
    spec = execnet.XSpec("ssh=%s" % sshaddress)
    spec.python = remotepython
    spec.ssh_config = ssh_config
    return execnet.default_group.makegateway(spec)

def APIWARN(version, msg, stacklevel=3):
    import warnings
    Warn = DeprecationWarning("(since version %s) %s" %(version, msg))
    warnings.warn(Warn, stacklevel=stacklevel)





def reporting(conn, result):
    log_map = {'debug': conn.logger.debug, 'error': conn.logger.error}
    while True:
        try:
            received = result.receive()
            level_received, message = list(received.items())[0]
            log_map[level_received](message.strip('\n'))
        except EOFError:
            break



from .log import reporting
from .util import admin_command


def _remote_run(channel, cmd):
    import subprocess
    import sys

    process = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    if process.stderr:
        while True:
            err = process.stderr.readline()
            if err == '' and process.poll() != None:
                break
            if err != '':
                channel.send({'error':err})
                sys.stderr.flush()
    if process.stdout:
        while True:
            out = process.stdout.readline()
            if out == '' and process.poll() != None:
                break
            if out != '':
                channel.send({'debug':out})
                sys.stdout.flush()


def run(conn, command, exit=False):
    """
    A real-time-logging implementation of a remote subprocess.Popen call where
    a command is just executed on the remote end and no other handling is done.
    """
    command = admin_command(conn.sudo, command)
    conn.logger.info('Running command: %s' % ' '.join(command))
    result = conn.execute(_remote_run, cmd=command)
    reporting(conn, result)
    if exit:
        conn.exit()


def _remote_check(channel, cmd):
    import subprocess

    process = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout = [line.strip('\n') for line in process.stdout.readlines()]
    stderr = [line.strip('\n') for line in process.stderr.readlines()]
    channel.send((stdout, stderr, process.wait()))


def check(conn, command, exit=False):
    """
    Execute a remote command with ``subprocess.Popen`` but report back the
    results in a tuple with three items: stdout, stderr, and exit status.

    This helper function *does not* provide any logging as it is the caller's
    responsibility to do so.
    """
    command = admin_command(conn.sudo, command)
    conn.logger.info('Running command: %s' % ' '.join(command))
    result = conn.execute(_remote_check, cmd=command)
    return result.receive()
    if exit:
        conn.exit()



"""
execnet io initialization code

creates io instances used for gateway io
"""
import os
import sys
from subprocess import Popen, PIPE

try:
    from execnet.gateway_base import Popen2IO, Message
except ImportError:
    from __main__ import Popen2IO, Message

class Popen2IOMaster(Popen2IO):
    def __init__(self, args):
        self.popen = p = Popen(args, stdin=PIPE, stdout=PIPE)
        Popen2IO.__init__(self, p.stdin, p.stdout)

    def wait(self):
        try:
            return self.popen.wait()
        except OSError:
            pass  # subprocess probably dead already

    def kill(self):
        killpopen(self.popen)

def killpopen(popen):
    try:
        if hasattr(popen, 'kill'):
            popen.kill()
        else:
            killpid(popen.pid)
    except EnvironmentError:
        sys.stderr.write("ERROR killing: %s\n" %(sys.exc_info()[1]))
        sys.stderr.flush()

def killpid(pid):
    if hasattr(os, 'kill'):
        os.kill(pid, 15)
    elif sys.platform == "win32" or getattr(os, '_name', None) == 'nt':
        try:
            import ctypes
        except ImportError:
            import subprocess
            # T: treekill, F: Force
            cmd = ("taskkill /T /F /PID %d" %(pid)).split()
            ret = subprocess.call(cmd)
            if ret != 0:
                raise EnvironmentError("taskkill returned %r" %(ret,))
        else:
            PROCESS_TERMINATE = 1
            handle = ctypes.windll.kernel32.OpenProcess(
                        PROCESS_TERMINATE, False, pid)
            ctypes.windll.kernel32.TerminateProcess(handle, -1)
            ctypes.windll.kernel32.CloseHandle(handle)
    else:
        raise EnvironmentError("no method to kill %s" %(pid,))



popen_bootstrapline = "import sys;exec(eval(sys.stdin.readline()))"


def popen_args(spec):
    python = spec.python or sys.executable
    args = [str(python), '-u']
    if spec is not None and spec.dont_write_bytecode:
        args.append("-B")
    # Slight gymnastics in ordering these arguments because CPython (as of
    # 2.7.1) ignores -B if you provide `python -c "something" -B`
    args.extend(['-c', popen_bootstrapline])
    return args

def ssh_args(spec):
    remotepython = spec.python or 'python'
    args = ['ssh', '-C' ]
    if spec.ssh_config is not None:
        args.extend(['-F', str(spec.ssh_config)])
    remotecmd = '%s -c "%s"' %(remotepython, popen_bootstrapline)
    args.extend([spec.ssh, remotecmd])
    return args



def create_io(spec):
    if spec.popen:
        args = popen_args(spec)
        return Popen2IOMaster(args)
    if spec.ssh:
        args = ssh_args(spec)
        io = Popen2IOMaster(args)
        io.remoteaddress = spec.ssh
        return io

RIO_KILL = 1
RIO_WAIT = 2
RIO_REMOTEADDRESS = 3
RIO_CLOSE_WRITE = 4

class RemoteIO(object):
    def __init__(self, master_channel):
        self.iochan = master_channel.gateway.newchannel()
        self.controlchan = master_channel.gateway.newchannel()
        master_channel.send((self.iochan, self.controlchan))
        self.io = self.iochan.makefile('r')


    def read(self, nbytes):
        return self.io.read(nbytes)

    def write(self, data):
        return self.iochan.send(data)

    def _controll(self, event):
        self.controlchan.send(event)
        return self.controlchan.receive()

    def close_write(self):
        self._controll(RIO_CLOSE_WRITE)

    def kill(self):
        self._controll(RIO_KILL)

    def wait(self):
        return self._controll(RIO_WAIT)

    def __repr__(self):
        return '<RemoteIO via %s>' % (self.iochan.gateway.id, )


def serve_remote_io(channel):
    class PseudoSpec(object):
        def __getattr__(self, name):
            return None
    spec = PseudoSpec()
    spec.__dict__.update(channel.receive())
    io = create_io(spec)
    io_chan, control_chan = channel.receive()
    io_target = io_chan.makefile()

    def iothread():
        initial = io.read(1)
        assert initial == '1'.encode('ascii')
        channel.gateway._trace('initializing transfer io for', spec.id)
        io_target.write(initial)
        while True:
            message = Message.from_io(io)
            message.to_io(io_target)
    import threading
    thread = threading.Thread(name='io-forward-'+spec.id,
                              target=iothread)
    thread.setDaemon(True)
    thread.start()

    def iocallback(data):
        io.write(data)
    io_chan.setcallback(iocallback)


    def controll(data):
        if data==RIO_WAIT:
            control_chan.send(io.wait())
        elif data==RIO_KILL:
            control_chan.send(io.kill())
        elif data==RIO_REMOTEADDRESS:
            control_chan.send(io.remoteaddress)
        elif data==RIO_CLOSE_WRITE:
            control_chan.send(io.close_write())
    control_chan.setcallback(controll)

if __name__ == "__channelexec__":
    serve_remote_io(channel)



"""
(c) 2008-2009, holger krekel
"""
import execnet

class XSpec:
    """ Execution Specification: key1=value1//key2=value2 ...
        * keys need to be unique within the specification scope
        * neither key nor value are allowed to contain "//"
        * keys are not allowed to contain "="
        * keys are not allowed to start with underscore
        * if no "=value" is given, assume a boolean True value
    """
    # XXX allow customization, for only allow specific key names
    popen = ssh = socket = python = chdir = nice = dont_write_bytecode = None

    def __init__(self, string):
        self._spec = string
        self.env = {}
        for keyvalue in string.split("//"):
            i = keyvalue.find("=")
            if i == -1:
                key, value = keyvalue, True
            else:
                key, value = keyvalue[:i], keyvalue[i+1:]
            if key[0] == "_":
                raise AttributeError("%r not a valid XSpec key" % key)
            if key in self.__dict__:
                raise ValueError("duplicate key: %r in %r" %(key, string))
            if key.startswith("env:"):
                self.env[key[4:]] = value
            else:
                setattr(self, key, value)

    def __getattr__(self, name):
        if name[0] == "_":
            raise AttributeError(name)
        return None

    def __repr__(self):
        return "<XSpec %r>" %(self._spec,)
    def __str__(self):
        return self._spec

    def __hash__(self):
        return hash(self._spec)
    def __eq__(self, other):
        return self._spec == getattr(other, '_spec', None)
    def __ne__(self, other):
        return self._spec != getattr(other, '_spec', None)

    def _samefilesystem(self):
        return bool(self.popen and not self.chdir)




from .lib import execnet


#
# Connection Object
#

class Connection(object):

    def __init__(self, hostname, logger=None, sudo=False):
        self.hostname = hostname
        self.gateway = execnet.makegateway('ssh=%s' % hostname)
        self.logger = logger or FakeRemoteLogger()
        self.sudo = sudo

    def execute(self, function, **kw):
        return self.gateway.remote_exec(function, **kw)

    def exit(self):
        self.gateway.exit()


class FakeRemoteLogger:

    def error(self, *a, **kw):
        pass

    def debug(self, *a, **kw):
        pass

    def info(self, *a, **kw):
        pass



"""
execnet: pure python lib for connecting to local and remote Python Interpreters.

(c) 2012, Holger Krekel and others
"""
__version__ = '1.1'

from . import apipkg

apipkg.initpkg(__name__, {
    'PopenGateway':     '.deprecated:PopenGateway',
    'SocketGateway':    '.deprecated:SocketGateway',
    'SshGateway':       '.deprecated:SshGateway',
    'makegateway':      '.multi:makegateway',
    'HostNotFound':     '.gateway_bootstrap:HostNotFound',
    'RemoteError':      '.gateway_base:RemoteError',
    'TimeoutError':     '.gateway_base:TimeoutError',
    'XSpec':            '.xspec:XSpec',
    'Group':            '.multi:Group',
    'MultiChannel':     '.multi:MultiChannel',
    'RSync':            '.rsync:RSync',
    'default_group':    '.multi:default_group',
    'dumps':            '.gateway_base:dumps',
    'loads':            '.gateway_base:loads',
    'DataFormatError':  '.gateway_base:DataFormatError',
})



"""
dispatching execution to threads

(c) 2009, holger krekel
"""
import threading
import time
import sys

# py2/py3 compatibility
try:
    import queue
except ImportError:
    import Queue as queue
if sys.version_info >= (3,0):
    exec ("def reraise(cls, val, tb): raise val")
else:
    exec ("def reraise(cls, val, tb): raise cls, val, tb")

ERRORMARKER = object()

class Reply(object):
    """ reply instances provide access to the result
        of a function execution that got dispatched
        through WorkerPool.dispatch()
    """
    _excinfo = None
    def __init__(self, task):
        self.task = task
        self._queue = queue.Queue()

    def _set(self, result):
        self._queue.put(result)

    def _setexcinfo(self, excinfo):
        self._excinfo = excinfo
        self._queue.put(ERRORMARKER)

    def get(self, timeout=None):
        """ get the result object from an asynchronous function execution.
            if the function execution raised an exception,
            then calling get() will reraise that exception
            including its traceback.
        """
        if self._queue is None:
            raise EOFError("reply has already been delivered")
        try:
            result = self._queue.get(timeout=timeout)
        except queue.Empty:
            raise IOError("timeout waiting for %r" %(self.task, ))
        if result is ERRORMARKER:
            self._queue = None
            excinfo = self._excinfo
            reraise(excinfo[0], excinfo[1], excinfo[2])
        return result

class WorkerThread(threading.Thread):
    def __init__(self, pool):
        threading.Thread.__init__(self)
        self._queue = queue.Queue()
        self._pool = pool
        self.setDaemon(1)

    def _run_once(self):
        reply = self._queue.get()
        if reply is SystemExit:
            return False
        assert self not in self._pool._ready
        task = reply.task
        try:
            func, args, kwargs = task
            result = func(*args, **kwargs)
        except (SystemExit, KeyboardInterrupt):
            return False
        except:
            reply._setexcinfo(sys.exc_info())
        else:
            reply._set(result)
        # at this point, reply, task and all other local variables go away
        return True

    def run(self):
        try:
            while self._run_once():
                self._pool._ready[self] = True
        finally:
            del self._pool._alive[self]
            try:
                del self._pool._ready[self]
            except KeyError:
                pass

    def send(self, task):
        reply = Reply(task)
        self._queue.put(reply)
        return reply

    def stop(self):
        self._queue.put(SystemExit)

class WorkerPool(object):
    """ A WorkerPool allows to dispatch function executions
        to threads.  Each Worker Thread is reused for multiple
        function executions. The dispatching operation
        takes care to create and dispatch to existing
        threads.

        You need to call shutdown() to signal
        the WorkerThreads to terminate and join()
        in order to wait until all worker threads
        have terminated.
    """
    _shuttingdown = False
    def __init__(self, maxthreads=None):
        """ init WorkerPool instance which may
            create up to `maxthreads` worker threads.
        """
        self.maxthreads = maxthreads
        self._ready = {}
        self._alive = {}

    def dispatch(self, func, *args, **kwargs):
        """ return Reply object for the asynchronous dispatch
            of the given func(*args, **kwargs) in a
            separate worker thread.
        """
        if self._shuttingdown:
            raise IOError("WorkerPool is already shutting down")
        try:
            thread, _ = self._ready.popitem()
        except KeyError: # pop from empty list
            if self.maxthreads and len(self._alive) >= self.maxthreads:
                raise IOError("can't create more than %d threads." %
                              (self.maxthreads,))
            thread = self._newthread()
        return thread.send((func, args, kwargs))

    def _newthread(self):
        thread = WorkerThread(self)
        self._alive[thread] = True
        thread.start()
        return thread

    def shutdown(self):
        """ signal all worker threads to terminate.
            call join() to wait until all threads termination.
        """
        if not self._shuttingdown:
            self._shuttingdown = True
            for t in list(self._alive):
                t.stop()

    def join(self, timeout=None):
        """ wait until all worker threads have terminated. """
        current = threading.currentThread()
        deadline = delta = None
        if timeout is not None:
            deadline = time.time() + timeout
        for thread in list(self._alive):
            if deadline:
                delta = deadline - time.time()
                if delta <= 0:
                    raise IOError("timeout while joining threads")
            thread.join(timeout=delta)
            if thread.isAlive():
                raise IOError("timeout while joining threads")

if __name__ == '__channelexec__':
    maxthreads = channel.receive()
    execpool = WorkerPool(maxthreads=maxthreads)
    gw = channel.gateway
    channel.send("ok")
    gw._trace("instantiated thread work pool maxthreads=%s" %(maxthreads,))
    while 1:
        gw._trace("waiting for new exec task")
        task = gw._execqueue.get()
        if task is None:
            gw._trace("thread-dispatcher got None, exiting")
            execpool.shutdown()
            execpool.join()
            raise gw._StopExecLoop
        gw._trace("dispatching exec task to thread pool")
        execpool.dispatch(gw.executetask, task)



import re

module_file = open("remoto/__init__.py").read()
metadata = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", module_file))
long_description = open('README.rst').read()

from setuptools import setup, find_packages

setup(
    name = 'remoto',
    description = 'Execute remote commands or processes.',
    packages = find_packages(),
    author = 'Alfredo Deza',
    author_email = 'contact [at] deza.pe',
    version = metadata['version'],
    url = 'http://github.com/alfredodeza/remoto',
    license = "MIT",
    zip_safe = False,
    keywords = "remote, commands, unix, ssh, socket, execute, terminal",
    long_description = long_description,
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Utilities',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
    ]
)
